@@ -7,6 +7,19 @@ use serde::Deserialize;
 use std::panic;
 use std::slice::from_ref;
 
+// Mimalloc purge function - forces allocator to return memory to OS
+extern "C" {
+    fn mi_collect(force: bool);
+}
+
+fn force_memory_release() {
+    unsafe {
+        for _ in 0..3 {
+            mi_collect(true);
+        }
+    }
+}
+
 #[derive(Deserialize)]
 pub struct GensonKwargs {
     #[serde(default = "default_ignore_outer_array")]
@@ -76,6 +89,12 @@ pub struct GensonKwargs {
 
     #[serde(default = "default_no_root_map")]
     pub no_root_map: bool,
+
+    /// Maximum number of schema builders to create in parallel at once.
+    /// Lower values reduce peak memory usage during schema inference.
+    /// If None, processes all strings at once. Default is None.
+    #[serde(default)]
+    pub max_builders: Option<usize>,
 }
 
 fn default_map_threshold() -> usize {
@@ -173,7 +192,7 @@ pub fn infer_json_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsResul
         // Original behavior: merge all schemas into one
         // We only need a single row, and we are allowed to change the length
         // Wrap EVERYTHING in panic catching, including config creation
-        let result = panic::catch_unwind(|| -> Result<String, String> {
+        let result = panic::catch_unwind(move || -> Result<String, String> {
             let config = SchemaInferenceConfig {
                 ignore_outer_array: kwargs.ignore_outer_array,
                 delimiter: if kwargs.ndjson { Some(b'\n') } else { None },
@@ -187,6 +206,7 @@ pub fn infer_json_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsResul
                 avro: kwargs.avro,
                 wrap_root: wrap_root_field.clone(),
                 no_root_map: kwargs.no_root_map,
+                max_builders: kwargs.max_builders,
                 debug: kwargs.debug,
                 profile: kwargs.profile,
                 verbosity: kwargs.verbosity,
@@ -195,6 +215,8 @@ pub fn infer_json_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsResul
             let schema_result = infer_json_schema_from_strings(&json_strings, config)
                 .map_err(|e| format!("Genson error: {}", e))?;
 
+            drop(json_strings);
+
             serde_json::to_string_pretty(&schema_result.schema)
                 .map_err(|e| format!("JSON serialization error: {}", e))
         });
@@ -215,7 +237,7 @@ pub fn infer_json_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsResul
         }
     } else {
         // New behavior: infer schema for each row individually
-        let result = panic::catch_unwind(|| -> Result<Vec<serde_json::Value>, String> {
+        let result = panic::catch_unwind(move || -> Result<Vec<serde_json::Value>, String> {
             let mut individual_schemas = Vec::new();
             for json_str in &json_strings {
                 let config = SchemaInferenceConfig {
@@ -231,6 +253,7 @@ pub fn infer_json_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsResul
                     avro: kwargs.avro,
                     wrap_root: wrap_root_field.clone(),
                     no_root_map: kwargs.no_root_map,
+                    max_builders: kwargs.max_builders,
                     debug: kwargs.debug,
                     profile: kwargs.profile,
                     verbosity: kwargs.verbosity,
@@ -240,6 +263,7 @@ pub fn infer_json_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsResul
                     .map_err(|e| format!("Individual genson error: {}", e))?;
                 individual_schemas.push(single_result.schema);
             }
+            drop(json_strings);
             Ok(individual_schemas)
         });
 
@@ -304,7 +328,7 @@ pub fn infer_polars_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsRes
     let wrap_root_field = kwargs.wrap_root.clone();
 
     // Use genson to infer JSON schema, then convert to Polars schema fields
-    let result = panic::catch_unwind(|| -> Result<Vec<(String, String)>, String> {
+    let result = panic::catch_unwind(move || -> Result<Vec<(String, String)>, String> {
         let config = SchemaInferenceConfig {
             ignore_outer_array: kwargs.ignore_outer_array,
             delimiter: if kwargs.ndjson { Some(b'\n') } else { None },
@@ -318,6 +342,7 @@ pub fn infer_polars_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsRes
             avro: kwargs.avro,
             wrap_root: wrap_root_field,
             no_root_map: kwargs.no_root_map,
+            max_builders: kwargs.max_builders,
             debug: kwargs.debug,
             profile: kwargs.profile,
             verbosity: kwargs.verbosity,
@@ -326,6 +351,8 @@ pub fn infer_polars_schema(inputs: &[Series], kwargs: GensonKwargs) -> PolarsRes
         let schema_result = infer_json_schema_from_strings(&json_strings, config)
             .map_err(|e| format!("Genson error: {}", e))?;
 
+        drop(json_strings);
+
         let format = if kwargs.avro {
             SchemaFormat::Avro
         } else {
@@ -434,57 +461,66 @@ pub fn normalise_json(inputs: &[Series], kwargs: GensonKwargs) -> PolarsResult<S
         PolarsError::ComputeError("Expected string column for JSON normalisation".into())
     })?;
 
-    // Collect all JSON strings
-    let mut json_strings = Vec::new();
-    for s in string_chunked.iter().flatten() {
-        if !s.trim().is_empty() {
-            json_strings.push(s.to_string());
+    let out = {
+        // Collect all JSON strings
+        let mut json_strings = Vec::new();
+        for s in string_chunked.iter().flatten() {
+            if !s.trim().is_empty() {
+                json_strings.push(s.to_string());
+            }
         }
-    }
 
-    let wrap_root_field = kwargs.wrap_root.clone();
+        let wrap_root_field = kwargs.wrap_root.clone();
 
-    // Infer schema ONCE
-    let config = SchemaInferenceConfig {
-        ignore_outer_array: kwargs.ignore_outer_array,
-        delimiter: if kwargs.ndjson { Some(b'\n') } else { None },
-        schema_uri: kwargs.schema_uri.clone(),
-        map_threshold: kwargs.map_threshold,
-        map_max_required_keys: kwargs.map_max_required_keys,
-        unify_maps: kwargs.unify_maps,
-        no_unify: kwargs.no_unify.iter().cloned().collect(),
-        force_field_types: kwargs.force_field_types.clone(),
-        wrap_scalars: kwargs.wrap_scalars,
-        avro: true, // normalisation implies Avro
-        wrap_root: wrap_root_field.clone(),
-        no_root_map: kwargs.no_root_map,
-        debug: kwargs.debug,
-        profile: kwargs.profile,
-        verbosity: kwargs.verbosity,
+        // Infer schema ONCE
+        let config = SchemaInferenceConfig {
+            ignore_outer_array: kwargs.ignore_outer_array,
+            delimiter: if kwargs.ndjson { Some(b'\n') } else { None },
+            schema_uri: kwargs.schema_uri.clone(),
+            map_threshold: kwargs.map_threshold,
+            map_max_required_keys: kwargs.map_max_required_keys,
+            unify_maps: kwargs.unify_maps,
+            no_unify: kwargs.no_unify.iter().cloned().collect(),
+            force_field_types: kwargs.force_field_types.clone(),
+            wrap_scalars: kwargs.wrap_scalars,
+            avro: true, // normalisation implies Avro
+            wrap_root: wrap_root_field.clone(),
+            no_root_map: kwargs.no_root_map,
+            max_builders: kwargs.max_builders,
+            debug: kwargs.debug,
+            profile: kwargs.profile,
+            verbosity: kwargs.verbosity,
+        };
+
+        let schema_result = infer_json_schema_from_strings(&json_strings, config).map_err(|e| {
+            PolarsError::ComputeError(format!("Schema inference failed: {e}").into())
+        })?;
+
+        drop(json_strings);
+
+        let schema = &schema_result.schema;
+
+        // Parse each row and normalise
+        let cfg = NormaliseConfig {
+            empty_as_null: kwargs.empty_as_null,
+            coerce_string: kwargs.coerce_string,
+            map_encoding: kwargs.map_encoding,
+            wrap_root: wrap_root_field.clone(),
+        };
+
+        let mut out = Vec::with_capacity(string_chunked.len());
+        for s in string_chunked {
+            let val = s
+                .and_then(|st| serde_json::from_str::<serde_json::Value>(st).ok())
+                .unwrap_or(serde_json::Value::Null);
+
+            let normed = normalise_values(vec![val], schema, &cfg).pop().unwrap();
+            out.push(serde_json::to_string(&normed).unwrap());
+        }
+        out
     };
 
-    let schema_result = infer_json_schema_from_strings(&json_strings, config)
-        .map_err(|e| PolarsError::ComputeError(format!("Schema inference failed: {e}").into()))?;
-
-    let schema = &schema_result.schema;
-
-    // Parse each row and normalise
-    let cfg = NormaliseConfig {
-        empty_as_null: kwargs.empty_as_null,
-        coerce_string: kwargs.coerce_string,
-        map_encoding: kwargs.map_encoding,
-        wrap_root: wrap_root_field.clone(),
-    };
-
-    let mut out = Vec::with_capacity(string_chunked.len());
-    for s in string_chunked {
-        let val = s
-            .and_then(|st| serde_json::from_str::<serde_json::Value>(st).ok())
-            .unwrap_or(serde_json::Value::Null);
-
-        let normed = normalise_values(vec![val], schema, &cfg).pop().unwrap();
-        out.push(serde_json::to_string(&normed).unwrap());
-    }
+    force_memory_release();
 
     Ok(Series::new("normalised".into(), out))
 }
